package com.chelvc.framework.database.interceptor;

import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import com.baomidou.mybatisplus.core.config.GlobalConfig;
import com.baomidou.mybatisplus.core.metadata.TableInfo;
import com.chelvc.framework.base.util.SpringUtils;
import com.chelvc.framework.common.util.ObjectUtils;
import com.google.common.collect.Lists;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.CtNewMethod;
import javassist.LoaderClassPath;
import javassist.NotFoundException;
import lombok.NonNull;
import org.apache.ibatis.parsing.XNode;
import org.springframework.core.annotation.Order;

/**
 * Mybatis配置拦截器抽象实现类
 *
 * @author Woody
 * @date 2024/4/6
 */
public abstract class MybatisConfigureInterceptor {
    /**
     * 是否已初始化
     */
    private static boolean INITIALIZED;

    /**
     * 拦截器实例列表
     */
    private static List<MybatisConfigureInterceptor> INSTANCES = Collections.emptyList();

    /**
     * 初始化配置拦截器
     */
    public synchronized static void initialize() {
        if (!INITIALIZED) {
            try {
                ClassPool pool = ClassPool.getDefault();
                pool.appendClassPath(new LoaderClassPath(Thread.currentThread().getContextClassLoader()));
                initializeMapperListener(pool);
                initializeTableInfoListener(pool);
                INSTANCES = Lists.newArrayList(initializeInterceptorInstances());
            } catch (Exception e) {
                throw new RuntimeException(e);
            } finally {
                INITIALIZED = true;
            }
        }
    }

    /**
     * 初始化Mapper解析监听器
     *
     * @param pool 类对象池
     * @throws Exception 初始化异常
     */
    private static void initializeMapperListener(ClassPool pool) throws Exception {
        // 替换XMLMapperBuilder.parse方法逻辑，加入Mapper解析监听回调逻辑
        CtClass clazz = pool.get("org.apache.ibatis.builder.xml.XMLMapperBuilder");
        CtMethod method = clazz.getDeclaredMethod("parse");
        method.setBody(String.format("{\n" +
                "if (!this.configuration.isResourceLoaded(this.resource)) {\n" +
                "   %s node = this.parser.evalNode(\"/mapper\");\n" +
                "   %s.configure(node);\n" +
                "   this.configurationElement(node);\n" +
                "   this.configuration.addLoadedResource(this.resource);\n" +
                "   this.bindMapperForNamespace();\n" +
                "}\n" +
                "this.parsePendingResultMaps();\n" +
                "this.parsePendingCacheRefs();\n" +
                "this.parsePendingStatements();\n" +
                "}", XNode.class.getName(), MybatisConfigureInterceptor.class.getName()));
        clazz.toClass();
    }

    /**
     * 初始化数据模型表信息监听器
     *
     * @param pool 类对象池
     * @throws Exception 初始化异常
     */
    private static void initializeTableInfoListener(ClassPool pool) throws Exception {
        // 获取原始initTableFields方法
        CtMethod initTableFields = null;
        CtClass clazz = pool.get("com.baomidou.mybatisplus.core.metadata.TableInfoHelper");
        for (CtMethod method : clazz.getDeclaredMethods()) {
            if (method.getName().equals("initTableFields")) {
                initTableFields = method;
            }
        }
        if (initTableFields == null) {
            throw new NotFoundException("initTableFields");
        }

        // 重命名initTableFields方法名
        CtMethod copy = CtNewMethod.copy(initTableFields, clazz, null);
        copy.setName("doInitTableFields");
        clazz.removeMethod(initTableFields);
        clazz.addMethod(copy);

        // 替换initTableFields方法，加入表信息初始化回调逻辑
        CtMethod replace = CtNewMethod.make(String.format("private static void initTableFields(" +
                        "Class clazz, %s globalConfig, %s tableInfo, java.util.List excludeProperty) {\n" +
                        "doInitTableFields(clazz, globalConfig, tableInfo, excludeProperty);\n" +
                        "%s.configure(tableInfo);\n" +
                        "}",
                GlobalConfig.class.getName(), TableInfo.class.getName(), MybatisConfigureInterceptor.class.getName()),
                clazz
        );
        clazz.addMethod(replace);
        clazz.toClass();
    }

    /**
     * 初始化Mybatis配置拦截器实例
     *
     * @return Mybatis配置拦截器列表
     * @throws Exception 初始化异常
     */
    private static List<MybatisConfigureInterceptor> initializeInterceptorInstances() throws Exception {
        // 查找当前包目录下MybatisConfigureInterceptor实现
        List<Class<?>> classes = SpringUtils.lookupClasses(
                MybatisConfigureInterceptor.class.getPackage().getName(),
                clazz -> clazz != MybatisConfigureInterceptor.class
                        && MybatisConfigureInterceptor.class.isAssignableFrom(clazz)
        );

        // 基于@Order注解对拦截器对象排序
        classes.sort(Comparator.comparingInt(
                clazz -> ObjectUtils.ifNull(clazz.getAnnotation(Order.class), Order::value, () -> 0)
        ));

        // 初始化对象实例
        if (ObjectUtils.isEmpty(classes)) {
            return Collections.emptyList();
        }
        List<MybatisConfigureInterceptor> interceptors = Lists.newArrayListWithCapacity(classes.size());
        for (Class<?> clazz : classes) {
            interceptors.add((MybatisConfigureInterceptor) clazz.newInstance());
        }
        return interceptors;
    }

    /**
     * 配置Mapper XML节点
     *
     * @param mapper Mapper XML节点
     */
    public static void configure(@NonNull XNode mapper) {
        INSTANCES.forEach(interceptor -> interceptor.intercept(mapper));
    }

    /**
     * 配置数据模型表信息
     *
     * @param table 数据模型表信息
     */
    public static void configure(@NonNull TableInfo table) {
        INSTANCES.forEach(interceptor -> interceptor.intercept(table));
    }

    /**
     * 拦截Mapper配置
     *
     * @param mapper Mapper XML节点
     */
    public void intercept(XNode mapper) {
    }

    /**
     * 拦截数据模型配置
     *
     * @param table 数据模型表信息
     */
    public void intercept(TableInfo table) {
    }
}
